from datetime import date, datetime, timedelta
from otter.const import FORMAT_DATE, FORMAT_TIME, FORMAT_DATETIME


# before 5 am it's still considered last work day
from otter.errors import FormatDatetimeError


def work_date(_date=date.today()):
    if datetime.now().hour < 5:
        return _date-timedelta(days=1)
    return _date


def yesterday():
    return date.today()-timedelta(days=1)


def date_string_to_date(obj):
    if obj.upper() == "TODAY":
        return date.today()
    if obj.upper() == "YESTERDAY":
        return yesterday()
    try:
        return datetime.strptime(obj, FORMAT_DATE).date()
    except ValueError:
        raise FormatDatetimeError(obj)


def date_to_date_string(obj):
    return obj.strftime(FORMAT_DATE)


def date_to_weekday_string(obj):
    return obj.strftime('%A')


def time_string_to_datetime(obj):
    if obj.upper() == "NOW":
        return datetime.now()
    try:
        return datetime.strptime(obj, FORMAT_DATETIME)
    except ValueError:
        try:
            today = date_to_date_string(date.today())
            return datetime.strptime(f"{today} {obj}", FORMAT_DATETIME)
        except ValueError:
            raise FormatDatetimeError(obj)


# TODO: Print full date time if not same work date
def datetime_to_time_string(obj):
    return obj.strftime(FORMAT_TIME)



import argparse
import sys

import otter.funcs as funcs
from otter.const import *
from otter.errors import *


try:
    from colorama import Fore, Back, Style, init
    init()
except ImportError:  # fallback so that the imported classes always exist
    class ColorFallback:
        __getattr__ = lambda self, name: ''
    Fore = Back = Style = ColorFallback()


def main(argv=None):
    try:
        main_internal(argv)
    except Error as e:
        print(e.message)
        sys.exit(1)
    except KeyboardInterrupt:
        print()  # print new line
        sys.exit(0)


def main_internal(argv):
    parser = argparse.ArgumentParser(description=OTTER_DESCRIPTION)
    parser.set_defaults(func=lambda x: parser.print_help())
    parser.add_argument('-v', '--version', action='version', version='%(prog)s ' + OTTER_VERSION)
    commands = parser.add_subparsers(metavar='commands')

    subcommand = commands.add_parser('start', help="Start time tracking for a new task")
    subcommand.add_argument('--time', nargs="?", const="NOW", default="NOW")
    subcommand.set_defaults(func=funcs.start)

    subcommand = commands.add_parser('stop', help="Stop time tracking for current task")
    subcommand.add_argument('--time', nargs="?", const="NOW", default="NOW")
    subcommand.set_defaults(func=funcs.stop)

    subcommand = commands.add_parser('describe', help="Describe work for current task")
    subcommand.add_argument('description')
    subcommand.set_defaults(func=funcs.describe)

    subcommand = commands.add_parser('status', help="Get current time tracking status")
    subcommand.set_defaults(func=funcs.status)

    subcommand = commands.add_parser('show', help="Show tracked time for given date")
    subcommand.add_argument('--date', nargs="?", const="TODAY", default="TODAY")
    subcommand.set_defaults(func=funcs.show)

    subcommand = commands.add_parser('sync', help="Sync from and to Odoo")
    subcommand.add_argument('--date', nargs="?", const="YESTERDAY", default="YESTERDAY")
    subcommand.set_defaults(func=funcs.sync)

    subcommand = commands.add_parser('login', help="Login to Odoo")
    subcommand.add_argument('--url')
    subcommand.add_argument('--database')
    subcommand.add_argument('--username')
    subcommand.add_argument('--password')
    subcommand.set_defaults(func=funcs.login)

    subcommand = commands.add_parser('logout', help="Logout from Odoo")
    subcommand.set_defaults(func=funcs.logout)

    try:
        import argcomplete
        argcomplete.autocomplete(parser)
    except:
        pass

    args = parser.parse_args(args=argv)
    args.func(args)


if __name__ == '__main__':
    main()



import json

from collections.abc import Iterable


class OtterEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        elif getattr(obj, "values", None):
            return obj.values()
        return json.JSONEncoder.default(self, obj)


def to_json(items):
    return json.dumps(items, indent=4, cls=OtterEncoder)


def from_json(string):
    return json.loads(string)


def serialize(args, items):
    if items is None:
        return None
    if isinstance(items, str):
        return items
    if not isinstance(items, Iterable) or not items:
        raise Exception('data cannot be serialized')

    # items = sorted(items)
    # if args.format == 'json':
    return to_json(items)
    # else:
    #     return '\n'.join(str(i) for i in items)


def serialize_file(file, items):
    with open(file, 'w') as f:
        json.dump(items, f, indent=2)


def deserialize_file(file):
    with open(file, 'r') as f:
        string = f.read()

        if len(string) == 0:
            return dict()

        if file.lower().endswith(".json"):  # args.format == 'json' or
            return from_json(string)

    return None



from datetime import datetime

from sqlalchemy.exc import NoResultFound

import otter.database.query as db
from otter import database
from otter.database.model import Project, ProjectTask, Record
from otter.model import OtterRecordSyncCalculator
from otter.odoo import rest
from otter.util.date import work_date


def get_or_create_record(project_task_id):
    """
    :param project_task_id: project task ID
    :return: record
    """

    current_date = work_date()
    session = database.session()

    try:
        return (
            session.query(
                Record
            )
            .filter(Record.date == current_date)
            .filter(Record.project_task_id == project_task_id)
            .filter(Record.odoo_id.is_(None))  # only unsynced
            .one()
        )
    except NoResultFound:
        record = db.add_record(project_task_id, session=session, add_date=current_date)
        session.commit()
        session.refresh(record)
        return record


def get_records(query_date, session=database.session()):
    return (
        session.query(
            Record
        )
        .filter(Record.date == query_date)
        .all()
    )


def sync_remote_projects(session=database.session()):
    projects_json = rest.get_projects_json()
    if projects_json is not None:
        for project_json in projects_json:
            session.merge(Project.from_odoo_json(project_json))
    session.commit()


def sync_remote_project_tasks(session=database.session()):
    project_tasks_json = rest.get_project_tasks_json()
    if project_tasks_json is not None:
        for task_json in project_tasks_json:
            session.merge(ProjectTask.from_odoo_json(task_json))
    session.commit()


def sync_local_records(query_date, session=database.session()):
    records = []  # get_records_until_date(date=query_date, only_unsynced=True, session)

    if len(records) > 0:
        record_sync_calc = OtterRecordSyncCalculator(records)

        for i, day in enumerate(record_sync_calc.dates()):
            # print() if i > 0 else None
            # print(f"{date_to_weekday_string(day)}, {date_to_date_string(day)}:")
            sync_records = record_sync_calc.records(day)

            for sync_record in sync_records:
                assert sync_record.odoo_id is None

                if sync_record.duration == 0:
                    session.delete(sync_record)
                    session.commit()
                    continue

                record_odoo_id = rest.post_record(sync_record.odoo_create_json())
                assert record_odoo_id is not None
                sync_record.odoo_id = record_odoo_id
                session.commit()


def sync_remote_records(session=database.session()):
    record_json = rest.get_records_json()
    if record_json is not None:
        for record_json in record_json:
            local_id = Record.get_id_by_odoo_id(record_json['id'], session)
            session.merge(Record.from_odoo_json(local_id, record_json))
    session.commit()



import getpass
import os
import sys

from otter import const, eprint


def input_auth(args):
    username = args.username
    password = args.password

    if username is None:
        username = input("Username: ")
    if args.username is None or password is None:
        password = getpass.getpass(prompt='Password: ', stream=None)

    return username, password


def input_confirm(args):
    if args.batch:
        return True
    return input_yes_no('Are you sure to continue?')


def input_yes_no(message, default='n'):
    choices = 'Y/n' if default.lower() == 'y' else 'y/N'
    while True:
        choice = input(f"{message} [{choices}]]: ").strip().lower()
        if choice not in ('', 'y', 'n'):
            continue
        if choice == '':
            choice = default
        return choice == 'y'


def input_default(key, default=None):
    if default is not None:
        choice = input("{} [{}]: ".format(key, default))
        return choice if choice != "" else default
    return input("{}: ".format(key))


def input_select_list_default(key, list, default=None):
    if list is None or len(list) == 0:
        return None

    if default is not None:
        print(f"{key} [{default}]:")
    else:
        print(f"{key}:")

    for i, item in enumerate(list):
        print(f"{i + 1}: {item}")

    choice = input("?: ")
    if default is not None:
        choice = choice if choice != "" else default
        if choice == default:
            return default

    return list[int(choice)-1]

def print_task_start(message):
    print(message, end=' ', flush=True)


def print_task_done(message="Done."):
    print(message)


def print_progress(title, percentage):
    sys.stdout.write("\r%s: %d%%" % (title, percentage))
    sys.stdout.flush()
    if percentage == 100:
        print()


def project_task_abbrev(project, task, offset_len=0):
    column_len = os.get_terminal_size().columns
    extra_len = offset_len + len(const.SEPARATOR)
    project_len = len(project)
    task_len = len(task)

    min_width = extra_len + 16 # 16 is min length for project + task

    # Require minimum 8 chars for both project and task
    if column_len < min_width: 
        eprint(f"Terminal too small, minimum width is {min_width} columns.")
        sys.exit(1)

    common_len = project_len + task_len
    target_len = min(common_len, column_len-extra_len)

    project_part = round(target_len * (float(project_len) / float(common_len)))
    task_part = target_len - project_part

    if project_len > project_part:
        project_list = list(project[:project_part])
        project_list[project_part-1] = const.DOTS
        project = ''.join(project_list)

    if task_len > task_part:
        task_list = list(task[:task_part])
        task_list[task_part-1] = const.DOTS
        task = ''.join(task_list)

    return f"{project}{const.SEPARATOR}{task}"



from otter.const import FORMAT_DATETIME, FORMAT_TIME


class Error(Exception):
    def __init__(self, message):
        self.message = message


class InconsistentDataError(Error):
    def __init__(self):
        super().__init__("Data is inconsistent, check the README for options.")


class InvalidChoiceError(Error):
    def __init__(self):
        super().__init__("Invalid choice.")


class NoProjectTasksFoundError(Error):
    def __init__(self):
        super().__init__("No project tasks found, run 'otter sync' to sync them.")


class NoRecordActiveError(Error):
    def __init__(self):
        super().__init__("No active task, use 'otter start' to start a task.")


class NoRecordForDateError(Error):
    def __init__(self, date):
        super().__init__("No task found for date '{}'.".format(date))


class FormatDatetimeError(Error):
    def __init__(self, datetime):
        super().__init__("Datetime '{}' cannot be matched against format '{}' or '{}'"
                         .format(datetime, FORMAT_DATETIME, FORMAT_TIME))



# based on https://realpython.com/python-sqlite-sqlalchemy/#using-flat-files-for-data-storage

from datetime import date

import otter.database as database
from otter.database.model import *


def add_record(project_task_id, session=database.session(), add_date=date.today):
    record = Record(project_task_id=project_task_id, date=add_date)
    session.add(record)
    return record


def add_record_part(name, record_id, session=database.session()):
    record_track_item = RecordTrackItem(name=name, record_id=record_id)
    session.add(record_track_item)
    return record_track_item


def add_record_time_slice(record_id, time_start=datetime.now(), session=database.session()):
    record_time_slice = RecordTimeSlice(time_start=time_start, record_id=record_id)
    session.add(record_time_slice)
    return record_time_slice


def query_records_most_recent(limit, session=database.session()):
    subquery = (
        session.query(
            Record.id,
            func.max(Record.date)
        )
        .group_by(Record.project_task_id)
        .subquery()
    )

    return (
        session.query(
            Record
        )
        .join(subquery, Record.id == subquery.c.id)
        .order_by(Record.date.desc())
        .limit(limit)
        .all()
    )


def query_record_time_slice_active(session=database.session()):
    return (
        session.query(
            RecordTimeSlice
        )
        .filter(RecordTimeSlice.time_stop.is_(None))  # only unstopped
        .one()
    )

def query_record_active(session=database.session()):
    """

    :param session:
    :return:
    """
    record_time_slice = query_record_time_slice_active(session=session)
    return record_time_slice.record



THIRD = 1.0/3
import math


def float_mod(num):
    return (int(num * 1000) % 1000) / 1000.0


def ceil_third(num):
    if float_mod(num) < THIRD:
        return math.floor(num)
    return math.ceil(num)


def round_duration(num):
    return ceil_third(num * 4.0) / 4.0


def ceil_duration(num):
    return math.ceil(num * 4.0) / 4.0


def floor_duration(num):
    return math.floor(num * 4.0) / 4.0
